import numpy as np
import pandas as pd



class SaveEndpointFromInset(SaveEndpoint):
    def __init__(self, save_file=None, output_filename='InsetChart.json', years_to_include=4):
        super().__init__(save_file=save_file, output_filename=output_filename)

        self.years_to_include = years_to_include


    def select_simulation_data(self, data, simulation):
        y = np.array([])
        cases = np.array([])
        infections = np.array([])
        EIR = np.array([])
        RDT_prev = np.array([])
        true_prev = np.array([])
        days_at_zero_true_prev = np.array([])
        true_prev_at_end = np.array([])
        avg_daily_bite_rate = np.array([])

        start_index = -(self.years_to_include*365+1)

        # Get last years of data:
        for j in range(self.years_to_include):
            s = j*365 + start_index
            e = (j+1)*365 + start_index
            cases_array = np.array(data[self.filenames[0]]["Channels"]["New Clinical Cases"]["Data"][s:e])
            infec_array = np.array(data[self.filenames[0]]["Channels"]["New Infections"]["Data"][s:e])
            EIR_array = np.array(data[self.filenames[0]]["Channels"]["Daily EIR"]["Data"][s:e])
            RDT_prev_arr = np.array(data[self.filenames[0]]["Channels"]["Blood Smear Parasite Prevalence"]["Data"][s:e])
            true_prev_arr = np.array(data[self.filenames[0]]["Channels"]["True Prevalence"]["Data"][s:e])
            daily_bite_rate = np.array(data[self.filenames[0]]["Channels"]["Daily Bites per Human"]["Data"][s:e])

            y = np.append(y,j)
            cases = np.append(cases, np.sum(cases_array))
            infections = np.append(infections, np.sum(infec_array))
            EIR = np.append(EIR, np.sum(EIR_array))
            RDT_prev = np.append(RDT_prev, np.average(RDT_prev_arr))
            true_prev = np.append(true_prev, np.average(true_prev_arr))
            days_at_zero_true_prev = np.append(days_at_zero_true_prev, np.sum(true_prev_arr[-1] == 0))
            true_prev_at_end = np.append(true_prev_at_end, true_prev_arr[-1])
            avg_daily_bite_rate = np.append(avg_daily_bite_rate, np.average(daily_bite_rate))

        sim_data = {
            "year": y,
            "cases": cases,
            "infections": infections,
            "EIR": EIR,
            "avg_RDT_prev": RDT_prev,
            "avg_true_prev": true_prev,
            "days_at_zero_true_prev": days_at_zero_true_prev,
            "true_prev_at_end": true_prev_at_end,
            "avg_daily_biting_rate": avg_daily_bite_rate
        }

        sim_data["sim_id"] = simulation.id
        for tag in simulation.tags:
            sim_data[tag] = simulation.tags[tag]

        return pd.DataFrame(sim_data)




class MinimumTransmissionFromInset(SaveEndpoint):
    def __init__(self, save_file=None, output_filename="InsetChart.json"):
        super().__init__(save_file=save_file, output_filename=output_filename)

    def select_simulation_data(self, data, simulation):
        data_inset = data[self.filenames[0]]

        # reported_cases = np.array(np.array(data[self.filenames[0]]["Channels"]["New Reported Cases"]["Data"])
        new_infections = np.array(data[self.filenames[0]]["Channels"]["New Infections"]["Data"])
        true_prevalence = np.array(data[self.filenames[0]]["Channels"]["True Prevalence"]["Data"])
        infected = np.array(data[self.filenames[0]]["Channels"]["Infected"]["Data"])
        d = np.arange(len(new_infections))

        # Elimination definition A: Far fewer actual new infections than importations
        # Note that according to Dan B, importations from the OutbreakIndividual campaign events would show up as "New Infections" in inset chart
        new_infections_in_last_year = np.zeros_like(new_infections)
        for i in np.arange(len(new_infections)):
            min_index = np.max([0,i-365])
            new_infections_in_last_year[i] = np.sum(new_infections[min_index:i])

        # Report minimum (after first year):
        new_infections_in_last_year_PADDED = np.copy(new_infections_in_last_year)
        new_infections_in_last_year_PADDED[:365] = np.max(new_infections_in_last_year_PADDED[365:])
        min_new_infections_index = np.argmin(new_infections_in_last_year_PADDED)
        min_new_infections_in_last_year = new_infections_in_last_year[min_new_infections_index]
        days_to_min_new_infections_in_last_year = d[min_new_infections_index]

        # Elimination definition B: true prevalence = 0
        min_true_prevalence_index = np.argmin(true_prevalence)
        min_true_prevalence = true_prevalence[min_true_prevalence_index]
        days_to_min_true_prevalence = d[min_true_prevalence_index]

        # Elimination definition C: infected = 0
        min_infected_index = np.argmin(infected)
        min_infected = true_prevalence[min_infected_index]
        days_to_min_infected = d[min_infected_index]

        sim_data = {
            "min_new_infections_in_last_year": [min_new_infections_in_last_year],
            "days_to_min_new_infections_in_last_year": [days_to_min_new_infections_in_last_year],
            "min_true_prevalence": [min_true_prevalence],
            "days_to_min_prevalence": [days_to_min_true_prevalence],
            "min_infected": [min_infected],
            "days_to_min_infected": [days_to_min_infected]
        }

        sim_data["sim_id"] = simulation.id
        for tag in simulation.tags:
            sim_data[tag] = simulation.tags[tag]

        return pd.DataFrame(sim_data)


def sum_channel_over_time_window(x, window_len=365):
    x_in_last_year = np.zeros_like(x)
    for i in np.arange(len(x)):
        min_index = np.max([0, i - window_len])
        x_in_last_year[i] = np.sum(x[min_index:i])
    return x_in_last_year


class MinimumTransmissionFromCounter(SaveEndpoint):
    def __init__(self, save_file=None, output_filename="ReportEventCounter.json"):
        super().__init__(save_file=save_file, output_filename=output_filename)


    def select_simulation_data(self, data, simulation):
        data_summary = data[self.filenames[0]]

        received_treatment = np.array(data_summary["Channels"]["Received_Treatment"]["Data"])
        # received_test = np.array(data_summary["Channels"]["Received_Test"]["Data"])
        received_RCD_drugs = np.array(data_summary["Channels"]["Received_RCD_Drugs"]["Data"])
        # received_campaign_drugs = np.array(data_summary["Channels"]["Received_Campaign_Drugs"]["Data"])
        d = np.arange(len(received_treatment))


        received_treatment_in_last_year = sum_channel_over_time_window(received_treatment)
        received_RCD_drugs = sum_channel_over_time_window(received_RCD_drugs)


        received_RCD_drugs_in_last_year = np.zeros_like(received_RCD_drugs)
        for i in np.arange(len(received_RCD_drugs)):
            min_index = np.max([0,i-365])
            received_RCD_drugs_in_last_year[i] = np.sum(received_RCD_drugs[min_index:i])

        # Elimination definition C:
        received_treatment_in_last_year_PADDED = np.copy(received_treatment_in_last_year)
        # received_test_in_last_year_PADDED = np.copy(received_test)
        received_RCD_drugs_in_last_year_PADDED = np.copy(received_RCD_drugs_in_last_year)
        # received_campaign_drugs_in_last_year_PADDED = np.copy(received_campaign_drugs)

        received_treatment_in_last_year_PADDED[:365] = np.max(received_treatment_in_last_year_PADDED[365:])
        # received_test_in_last_year_PADDED[:365] = np.max(received_test_in_last_year_PADDED[365:])
        received_RCD_drugs_in_last_year_PADDED[:365] = np.max(received_RCD_drugs_in_last_year_PADDED[365:])
        # received_campaign_drugs_in_last_year_PADDED[:365] = np.max(received_campaign_drugs_in_last_year_PADDED[365:])

        min_index = np.argmin(received_treatment_in_last_year_PADDED)
        min_received_treatment_in_last_year = received_treatment_in_last_year_PADDED[min_index]
        days_to_min_received_treatment_in_last_year = d[min_index]

        min_index = np.argmin(received_RCD_drugs_in_last_year_PADDED)
        min_received_RCD_drugs_in_last_year = received_RCD_drugs_in_last_year_PADDED[min_index]
        days_to_min_received_RCD_drugs_in_last_year = d[min_index]


        sim_data = {
            "min_received_treatment_in_last_year": [min_received_treatment_in_last_year],
            "days_to_min_received_treatment_in_last_year": [days_to_min_received_treatment_in_last_year],
            "min_received_RCD_drugs_in_last_year": [min_received_RCD_drugs_in_last_year],
            "days_to_min_received_RCD_drugs_in_last_year": [days_to_min_received_RCD_drugs_in_last_year]
        }

        sim_data["sim_id"] = simulation.id
        for tag in simulation.tags:
            sim_data[tag] = simulation.tags[tag]

        return pd.DataFrame(sim_data)




def run_elimination_analyzers_and_save_output(exp_id, include_counter=True):

    SetupParser.default_block = 'HPC'
    SetupParser.init()

    am = AnalyzeManager(force_analyze=False)

    analyze_summary = MinimumTransmissionFromInset(save_file=None)
    analyzer_list = [analyze_summary]

    if include_counter:
        analyze_counter = MinimumTransmissionFromCounter(save_file=None)
        analyzer_list += [analyze_counter]

    for analyzer in analyzer_list:
        am.add_analyzer(analyzer)
    exp = retrieve_experiment(exp_id)
    print(exp)
    am.add_experiment(exp)
    am.analyze()

    # Put results together into a single DataFrame:
    df_list = [x.results for x in am.analyzers]
    if len(df_list) == 1:
        df_result = df_list[0]
    elif len(df_list) == 2:
        print("flag1")
        df_result = pd.merge(df_list[0], df_list[1],
                             on="sim_id", suffixes=["","_duplicated"])
        print("flag2")
    else:
        raise NotImplementedError

    df_result = remove_duplicate_columns(df_result)

    df_result.to_csv("min_transmission_{}.csv".format(exp_id), index=False)
    return df_result





def save_biting_data(exp_id, years_to_include=1):
    SetupParser.default_block = 'HPC'
    SetupParser.init()

    am = AnalyzeManager(force_analyze=False)
    analyze_inset = SaveEndpointFromInset(output_filename="ReportMalariaFilteredFinal_Year.json",
                                          save_file=None,
                                          years_to_include=1)

    analyzer_list = [analyze_inset]

    for analyzer in analyzer_list:
        am.add_analyzer(analyzer)
    exp = retrieve_experiment(exp_id)
    print(exp)
    am.add_experiment(exp)
    am.analyze()

    df_result = am.analyzers[0].results
    df_result.to_csv("biting_{}.csv".format(exp_id), index=False)
    return df_result







## SPATIAL


def construct_spatial_output_df(rawdata, channel, timesteps=[]):
    n_nodes = rawdata['n_nodes']
    n_tstep = rawdata['n_tstep']
    if 'start' in rawdata:
        start = rawdata['start']
        interval = rawdata['interval']
    else :
        start, interval = 0,1
    nodeids = rawdata['nodeids']
    data = rawdata['data']

    all_timesteps = range(start, (start+n_tstep)*interval, interval)

    df = pd.DataFrame( { channel: [item for sublist in data for item in sublist],
                         'time': [item for sublist in [[x]*n_nodes for x in all_timesteps] for item in sublist],
                         'node': [item for sublist in [list(nodeids)*len(all_timesteps)] for item in sublist]} )
    if len(timesteps)==0:
        return df
    else:
        # timesteps = sorted(list(set(all_timesteps).intersection(timesteps)))
        # return df[df['time'].isin(timesteps)]
        return df[np.in1d(df['time'], timesteps)]


class GetSpatialBinaryAsDataframe(BaseAnalyzer):
    def __init__(self, save_file=None, channel="Population", timesteps=None, exclude_tags=["u5_hs_rate", "o5_hs_rate", "exe_collection_id", "input_collection_id"], save_only_nonzero_values=False):
        output_filename = "SpatialReportMalariaFiltered_{}.bin".format(channel)
        filenames = ['output/{}'.format(output_filename)]
        super().__init__(filenames=filenames)

        self.channel = channel
        self.timesteps = timesteps
        self.save_file = save_file
        self.exclude_tags = exclude_tags
        self.save_only_nonzero_values = save_only_nonzero_values


    def select_simulation_data(self, data, simulation):
        df = construct_spatial_output_df(data[self.filenames[0]], self.channel, timesteps=self.timesteps)

        if self.save_only_nonzero_values and self.channel == "Blood_Smear_Parasite_Prevalence":
            df = df[df[self.channel] > 0].reset_index(drop=True)

        # Add simulation information/tags
        df["sim_id"] = simulation.id
        for tag in simulation.tags:
            if tag not in self.exclude_tags:
                df[tag] = simulation.tags[tag]
        return df


    def combine(self, all_data):
        data_list = []
        for sim in all_data.keys():
            data_list.append(all_data[sim])

        return pd.concat(data_list).reset_index(drop=True)


    def finalize(self, all_data):
        sim_data_full = self.combine(all_data)
        if self.save_file:
            sim_data_full.to_csv(self.save_file, index=False)
        return sim_data_full


def run_analyzers_for_spatial_check(exp_id, save_only_nonzero_values=False):
    SetupParser.default_block = 'HPC'
    SetupParser.init()

    # debugging only
    import os
    print("CPU count: ", os.cpu_count())

    timesteps = np.arange(3*365, 4*365, 60)

    am = AnalyzeManager()
    # debugging only
    am.max_threads = min(os.cpu_count(), 48)

    analyzer_list = [GetSpatialBinaryAsDataframe(channel="Population", timesteps=timesteps, save_only_nonzero_values=save_only_nonzero_values),
                     GetSpatialBinaryAsDataframe(channel="Blood_Smear_Parasite_Prevalence", timesteps=timesteps, save_only_nonzero_values=save_only_nonzero_values)]
    # GetSpatialBinaryAsDataframe(channel="Daily_Bites_Per_Human", timesteps=timesteps)]
    for analyzer in analyzer_list:
        am.add_analyzer(analyzer)
    exp = retrieve_experiment(exp_id)
    am.add_experiment(exp)
    am.analyze()

    results = [x.results for x in am.analyzers]

    # Combine results
    df_return = pd.merge(results[0], results[1], how="inner")
    df_return.to_csv("spatial_{}.csv".format(exp_id), index=False)
    return df_return



# def find_most_representative_seed_of_experiment()
# run endpoint analyzer
# use output and find median