import numpy as np
import pandas as pd
from idmtools.entities import IAnalyzer


class SimEndpoint(IAnalyzer):
    def __init__(self,
                 save_file=None,
                 window_size=3*365,
                 include_inset=True,
                 include_counter=True,
                 include_summary=False,
                 only_keep_tags=None,
                 rdt_channel_as_hrp2=True
                 ):


        filenames = []
        if include_inset:
            filenames += ['output/InsetChart.json']
        if include_counter:
            filenames += ['output/ReportEventCounter.json']
        if include_summary:
            raise NotImplementedError
            # filenames += ['output/MalariaSummaryReport_AnnualAverage.json']

        if filenames == []:
            raise ValueError("Filenames cannot be empty")

        super().__init__(filenames=filenames)

        self.save_file = save_file
        self.window_size = window_size
        self.include_inset = include_inset
        self.include_counter = include_counter
        self.include_summary = include_summary
        self.only_keep_tags = only_keep_tags
        self.rdt_channel_as_hrp2 = rdt_channel_as_hrp2


    def map(self, data, simulation):
        sim_dict = {}

        if self.include_inset:
            inset_dict = data[self.filenames[0]]

            aeir = np.sum(inset_dict["Channels"]["Daily EIR"]["Data"][-self.window_size:])*365/self.window_size
            if self.rdt_channel_as_hrp2:
                annual_rdt_prev = np.mean(inset_dict["Channels"]["PfHRP2 Prevalence"]["Data"][-self.window_size:])
            else:
                annual_rdt_prev = np.mean(inset_dict["Channels"]["Blood Smear Parasite Prevalence"]["Data"][-self.window_size:])
            pfemp_frac = np.mean(inset_dict["Channels"]["Variant Fraction-PfEMP1 Major"]["Data"][-self.window_size:])
            pop = np.mean(inset_dict["Channels"]["Statistical Population"]["Data"][-self.window_size:])
            annual_cases_per_1000 = np.sum(inset_dict["Channels"]["New Clinical Cases"]["Data"][-self.window_size:])*365/self.window_size * 1000/pop

            sim_dict["aeir"] = aeir
            sim_dict["annual_cases_per_1000"] = annual_cases_per_1000
            sim_dict["annual_rdt_prev"] = annual_rdt_prev
            sim_dict["pfemp_frac"] = pfemp_frac

        if self.include_counter:
            if not self.include_inset:
                raise NotImplementedError("Need pop from InsetChart")

            counter_dict = data[self.filenames[1]]
            if "Received_Treatment" in counter_dict["Channels"]:
                annual_reported_cases_per_1000 = np.sum(counter_dict["Channels"]["Received_Treatment"]["Data"][-self.window_size:])*365/self.window_size * 1000/pop
                sim_dict["annual_reported_cases_per_1000"] = annual_reported_cases_per_1000

        if self.include_summary:
            raise NotImplementedError

        # Prepare the sim data for returning
        sim_data = pd.DataFrame(sim_dict, index=[0])

        sim_data["sim_id"] = simulation.id
        if self.only_keep_tags is None:
            for tag in simulation.tags:
                sim_data[tag] = simulation.tags[tag]
        else:
            for tag in simulation.tags:
                if tag in self.only_keep_tags:
                    sim_data[tag] = simulation.tags[tag]

        return sim_data


    def combine(self, all_data):
        data_list = []
        for sim in all_data.keys():
            data_list.append(all_data[sim])

        if len(data_list) == 1:
            return data_list[0]

        return pd.concat(data_list, ignore_index=True).reset_index(drop=True)


    def reduce(self, all_data):
        sim_data_full = self.combine(all_data)

        if self.save_file:
            sim_data_full.to_csv(self.save_file, index=False)
        return sim_data_full
