import numpy as np
import pandas as pd


def get_pop_weighted_channel_from_summary_report(data_summary, channel, num_years):
    pop_by_age_bin = data_summary['DataByTimeAndAgeBins']['Average Population by Age Bin']
    channel_of_interest = data_summary['DataByTimeAndAgeBins'][channel]

    pop_weighted_channel = np.array([
        np.sum(np.array(channel_of_interest[i]) * np.array(pop_by_age_bin[i])) for i in range(num_years)
    ])

    return pop_weighted_channel


def get_population_from_summary_report(data_summary, num_years):
    pop_by_age_bin = data_summary['DataByTimeAndAgeBins']['Average Population by Age Bin']

    pop = np.array([
        np.sum(np.array(pop_by_age_bin[i])) for i in range(num_years)
    ])

    return pop


def get_annual_cases_from_summary_report(data_summary, num_years):
    return get_pop_weighted_channel_from_summary_report(data_summary, 'Annual Clinical Incidence by Age Bin', num_years)


def get_severe_cases_from_summary_report(data_summary, num_years):
    return get_pop_weighted_channel_from_summary_report(data_summary, 'Annual Severe Incidence by Age Bin', num_years)


def get_annual_EIR_from_summary_report(data_summary, num_years, num_nodes=30):
    aeir = np.array(data_summary['DataByTime']['Annual EIR'][:num_years])

    print("Correcting since Summary Report incorrectly adds EIR across nodes...")
    aeir = aeir/num_nodes

    return aeir


def get_annual_avg_RDT_prev_from_summary_report(data_summary, num_years):
    pop_by_age_bin = data_summary['DataByTimeAndAgeBins']['Average Population by Age Bin']
    rdtprev_by_age_bin = data_summary['DataByTimeAndAgeBins']['PfPR by Age Bin']

    rdt_prev = np.array([
        np.sum(np.array(rdtprev_by_age_bin[i]) * np.array(pop_by_age_bin[i])) / np.sum(pop_by_age_bin[i]) for i in range(num_years)
    ])
    return rdt_prev


def get_average_age_from_summary_report(data_summary, num_years, channel=None):
    # If channel is None, return average age of population.
    # Otherwise, return average age from that channel (e.g. Annual Severe Incidence by Age Bin)"
    age_bins = data_summary['Metadata']['Age Bins']

    if not channel:
        pop_by_age_bin = data_summary['DataByTimeAndAgeBins']['Average Population by Age Bin']
        pop_avg_age = np.array([
            np.sum(np.array(age_bins) * np.array(pop_by_age_bin[i])) /np.sum(pop_by_age_bin[i]) for i in range(num_years)
        ])
        return pop_avg_age
    else:
        incidence_by_age_bin = data_summary['DataByTimeAndAgeBins'][channel]
        incidence_avg_age = np.array([
            np.sum(np.array(age_bins) * np.array(incidence_by_age_bin[i])) /np.sum(incidence_by_age_bin[i]) for i in range(num_years)
        ])
        return incidence_avg_age


def get_average_age_population(data_summary, num_years):
    return get_average_age_from_summary_report(data_summary, num_years)


def get_average_age_clinical_cases(data_summary, num_years):
    return get_average_age_from_summary_report(data_summary, num_years,'Annual Clinical Incidence by Age Bin')


def get_average_age_severe_cases(data_summary, num_years):
    return get_average_age_from_summary_report(data_summary, num_years,'Annual Severe Incidence by Age Bin')



class SaveEndpoint(BaseAnalyzer):
    def __init__(self, save_file=None, output_filename="InsetChart.json"):
        filenames = ['output/{}'.format(output_filename)]
        super().__init__(filenames=filenames)

        self.save_file = save_file

    def combine(self, all_data):
        data_list = []
        for sim in all_data.keys():
            data_list.append(all_data[sim])

        return pd.concat(data_list).reset_index(drop=True)

    def finalize(self, all_data):
        sim_data_full = self.combine(all_data)
        if self.save_file:
            sim_data_full.to_csv(self.save_file, index=False)
        return sim_data_full


class SaveEndpointFromSummaryReport(SaveEndpoint):
    def __init__(self, save_file=None, output_filename="MalariaSummaryReport_AnnualAverage.json", years_to_include=4, num_nodes=30):
        super().__init__(save_file=save_file, output_filename=output_filename)

        self.years_to_include = years_to_include
        self.num_nodes = num_nodes


    def select_simulation_data(self, data, simulation):
        data_summary = data[self.filenames[0]]

        y = np.arange(self.years_to_include)
        cases = get_annual_cases_from_summary_report(data_summary, self.years_to_include)
        severe_cases = get_severe_cases_from_summary_report(data_summary, self.years_to_include)

        EIR = get_annual_EIR_from_summary_report(data_summary, self.years_to_include, self.num_nodes)
        RDT_prev = get_annual_avg_RDT_prev_from_summary_report(data_summary, self.years_to_include)

        average_age_population = get_average_age_population(data_summary, self.years_to_include)
        average_age_clinical = get_average_age_clinical_cases(data_summary, self.years_to_include)
        average_age_severe = get_average_age_severe_cases(data_summary, self.years_to_include)
        population = get_population_from_summary_report(data_summary, self.years_to_include)

        sim_data = {
            "year": y,
            "cases": cases,
            "severe_cases": severe_cases,
            "EIR": EIR,
            "avg_RDT_prev": RDT_prev,
            "average_age_population": average_age_population,
            "average_age_clinical": average_age_clinical,
            "average_age_severe": average_age_severe,
            "population": population
        }


        sim_data["sim_id"] = simulation.id
        for tag in simulation.tags:
            sim_data[tag] = simulation.tags[tag]

        return pd.DataFrame(sim_data)



