import numpy as np
import pandas as pd
import scipy.interpolate

def fit_univariate_spline(x, y, s=1e5, xmin=None, xmax=None, k=3):
    from scipy.interpolate import UnivariateSpline

    argsort_x = np.argsort(x)
    x = x[argsort_x]
    y = y[argsort_x]

    spl = UnivariateSpline(x, y, s=s, k=k)

    x_range = np.max(x) - np.min(x)
    if not xmin:
        xmin = np.min(x) + 0.03 * x_range
    if not xmax:
        xmax = np.max(x) - 0.03 * x_range
    xs = np.linspace(xmin, xmax, 500)
    ys = spl(xs)
    return xs, ys


def fit_lowess_spline(x, y, frac=0.67, return_unique=False, return_error=False, percentile_top=80, percentile_bottom=20):
    # From https://james-brennan.github.io/posts/lowess_conf/
    from statsmodels.nonparametric.smoothers_lowess import lowess as sm_lowess

    def smooth(x, y, xgrid, frac=0.67):
        samples = np.random.choice(len(x), 50, replace=True)
        y_s = y[samples]
        x_s = x[samples]
        y_sm = sm_lowess(y_s, x_s, frac=frac, it=5,
                         return_sorted=False)
        # regularly sample it onto the grid
        y_grid = scipy.interpolate.interp1d(x_s, y_sm,
                                            fill_value='extrapolate')(xgrid)
        return y_grid

    if not return_error:
        sm_x, sm_y = sm_lowess(y, x, frac=frac, return_sorted=True).T
        if return_unique:
            sm_x_unique, unique_indices = np.unique(sm_x, return_index=True)
            sm_y_unique = sm_y[unique_indices]
            return sm_x_unique, sm_y_unique
        else:
            return sm_x, sm_y
    elif return_error:
        xgrid = np.linspace(x.min(), x.max())
        K = 100
        smooths = np.stack([smooth(x, y, xgrid, frac=frac) for k in range(K)]).T

        mean = np.nanmean(smooths, axis=1)
        #         median = np.nanmedian(smooths, axis=1)
        q_top = np.nanpercentile(smooths, percentile_top, axis=1)
        q_bot = np.nanpercentile(smooths, percentile_bottom, axis=1)

        return xgrid, mean, q_bot, q_top





def bin_dataframe_column(df, col, bin_edges):
    if isinstance(bin_edges, list):
        bin_edges = np.array(bin_edges)

    n_bins = len(bin_edges)-1

    df[col+"_BIN_NUMBER"] = -1

    bin_mean = np.array([])
    bin_median = np.array([])

    for bi in np.arange(n_bins):
        cut = np.logical_and(df[col] >= bin_edges[bi], df[col] < bin_edges[bi+1])
        df.loc[cut, col+"_BIN_NUMBER"] = int(bi)
        df.loc[cut, col+"_BIN_MIN"] = bin_edges[bi]
        df.loc[cut, col+"_BIN_MAX"] = bin_edges[bi+1]
        df.loc[cut, col+"_BIN_CENTER"] = 0.5*(bin_edges[bi] + bin_edges[bi+1])
        df.loc[cut, col+"_BIN_MEAN"] = np.mean(df[col][cut])
        df.loc[cut, col+"_BIN_MEDIAN"] = np.median(df[col][cut])

        bin_mean = np.append(bin_mean, np.mean(df[col][cut]))
        bin_median = np.append(bin_median, np.median(df[col][cut]))

    bin_df = pd.DataFrame({
        col + "_BIN_NUMBER": list(range(n_bins)),
        col + "_BIN_MIN": bin_edges[:-1],
        col + "_BIN_MAX": bin_edges[1:],
        col + "_BIN_CENTER": 0.5 * (bin_edges[:-1] + bin_edges[1:]),
        col + "_BIN_MEAN": bin_mean,
        col + "_BIN_MEDIAN": bin_median
    })

    return df, bin_df

def bin_column1_and_return_aggregated_column2(df, col1, col2, col1_bin_edges, agg_col2="median_and_spread", **kwargs):
    df, bin_df = bin_dataframe_column(df, col1, col1_bin_edges)

    if agg_col2 in ["mean", "median", "std"]:
        col2_binned_df = df.groupby(col1 + "_BIN_NUMBER").agg({col2: agg_col2}).reset_index()
        count = df.groupby(col1 + "_BIN_NUMBER").agg({col2: "count"}).reset_index().rename(columns={col2: col2 + "_COUNT"})

        col2_binned_df = col2_binned_df.merge(count)

    elif agg_col2 == "median_and_spread":

        # Oldest method
        med = df.groupby(col1 + "_BIN_NUMBER").agg({col2: "median"}).reset_index()
        Q1 = df.groupby(col1 + "_BIN_NUMBER").agg({col2: lambda x: np.percentile(x,20)}).reset_index().rename(columns={col2: col2 + "_Q1"})
        Q3 = df.groupby(col1 + "_BIN_NUMBER").agg({col2: lambda x: np.percentile(x,80)}).reset_index().rename(columns={col2: col2 + "_Q3"})
        count = df.groupby(col1 + "_BIN_NUMBER").agg({col2: "count"}).reset_index().rename(columns={col2: col2 + "_COUNT"})
        col2_binned_df = pd.merge(med, Q1)
        col2_binned_df = col2_binned_df.merge(Q3)
        col2_binned_df = col2_binned_df.merge(count)

        # updated: https://stackoverflow.com/questions/12589481/multiple-aggregations-of-the-same-column-using-pandas-groupby-agg
        # agg_dict = {
        #     "{}_MEDIAN".format(col2): np.median,
        #     "{}_PERCENTILE20".format(col2): lambda x: np.percentile(x,20),
        #     "{}_PERCENTILE80".format(col2): lambda x: np.percentile(x,80),
        # }
        # df["{}_p20".format(col2)] = df[col2]
        # df["{}_p80".format(col2)] = df[col2]
        # agg_dict = {
        #     col2: np.median,
        #     "{}_p20".format(col2): lambda x: np.percentile(x,20),
        #     "{}_p80".format(col2): lambda x: np.percentile(x,80),
        # }
        # col2_binned_df = df.groupby(col1 + "_BIN_NUMBER").agg({"col2": agg_dict})


        # Cleaner, but we are on old Pandas...: https://pandas-docs.github.io/pandas-docs-travis/user_guide/groupby.html#named-aggregation
        # col2_binned_df = df.groupby(col1 + "_BIN_NUMBER").agg(
        #     median=pd.NamedAgg(column=col2, aggfunc="median"),
        #     percentile_20=pd.NamedAgg(column=col2, aggfunc=lambda x: np.percentile(x,20)),
        #     percentile_80=pd.NamedAgg(column=col2, aggfunc=lambda x: np.percentile(x,80))
        # )

    else:
        raise NotImplementedError

    print(col2_binned_df)
    col2_binned_df = col2_binned_df.merge(bin_df, how="left", on = col1 + "_BIN_NUMBER")

    return col2_binned_df