import numpy as np
import pandas as pd
from idmtools.entities import IAnalyzer


class BurninEndpoint(IAnalyzer):
    def __init__(self, save_file=None, window_size=3*365, only_keep_tags=None):
        filenames = ['output/InsetChart.json', 'output/ReportEventCounter.json']
        super().__init__(filenames=filenames)
        self.save_file = save_file
        self.window_size = window_size
        self.only_keep_tags = only_keep_tags
        self.data_columns = ["aeir", "annual_cases_per_1000", "annual_rdt_prev", "annual_reported_cases_per_1000"]

    def map(self, data, simulation):
        inset_dict = data[self.filenames[0]]
        counter_dict = data[self.filenames[1]]

        pop = np.mean(inset_dict["Channels"]["Statistical Population"]["Data"][-self.window_size:])
        aeir = np.sum(inset_dict["Channels"]["Daily EIR"]["Data"][-self.window_size:])*365/self.window_size
        annual_cases_per_1000 = np.sum(inset_dict["Channels"]["New Clinical Cases"]["Data"][-self.window_size:])*365/self.window_size * 1000/pop
        annual_rdt_prev = np.mean(inset_dict["Channels"]["Blood Smear Parasite Prevalence"]["Data"][-self.window_size:])

        sim_dict = {
            "aeir": aeir,
            "annual_cases_per_1000": annual_cases_per_1000,
            "annual_rdt_prev": annual_rdt_prev
        }

        if "Received_Treatment" in counter_dict["Channels"]:
            annual_reported_cases_per_1000 = np.sum(counter_dict["Channels"]["Received_Treatment"]["Data"][-self.window_size:])*365/self.window_size * 1000/pop
            sim_dict["annual_reported_cases_per_1000"] = annual_reported_cases_per_1000
        else:
            raise NotImplementedError

        sim_data = pd.DataFrame(sim_dict, index=[0])

        sim_data["sim_id"] = simulation.id
        if self.only_keep_tags is None:
            for tag in simulation.tags:
                sim_data[tag] = simulation.tags[tag]
        else:
            for tag in simulation.tags:
                if tag in self.only_keep_tags:
                    sim_data[tag] = simulation.tags[tag]

        return sim_data


    # build in capability if you run multiple seeds, to average over them.
    def combine(self, all_data):
        data_list = []
        for sim in all_data.keys():
            data_list.append(all_data[sim])

        if len(data_list) == 1:
            return data_list[0]

        return pd.concat(data_list, ignore_index=True).reset_index(drop=True)


    def reduce(self, all_data):
        sim_data_full = self.combine(all_data)

        if "Run_Number" in sim_data_full.columns and len(list(set(sim_data_full["Run_Number"]))) > 1:
            if self.only_keep_tags is None:
                raise NotImplementedError
            else:
                cols_to_groupby = self.only_keep_tags

                agg_dict = {}
                for col in self.data_columns:
                    agg_dict[col] = "mean"  # could later try ["mean", "std"]
                return_data = sim_data_full.groupby(cols_to_groupby).agg(agg_dict).reset_index()
        else:
            return_data = sim_data_full

        if self.save_file:
            return_data.to_csv(self.save_file, index=False)
        return return_data
