import os.path
from copy import copy

import numpy as np
import pandas as pd
from idmtools.core.platform_factory import Platform
from idmtools.entities.experiment import Experiment
from matplotlib import pyplot as plt
import seaborn as sns

from jsuresh_helpers.analyzers.create_sim_map import get_sim_map
from jsuresh_helpers.burnins.run_ssmt_sharon_hack import run_burnin_endpoint_ssmt

sns.set_context("talk")

# from jsuresh_helpers.analyzers.run_analyzer import run_analyzer_locally, run_analyzer_as_ssmt
from jsuresh_helpers.binning_and_smoothing import fit_lowess_spline


#fixme Needs transmission tags to be changed if aggregating over multiple tags.
def find_representative_of_burnin_sweep(experiment_id, tags_to_group_by):
    run_burnin_endpoint_ssmt(experiment_id)
    survey_burnin_sweep(experiment_id)


def survey_burnin_sweep(df_burnin, experiment_id=None, tags_to_group_by="funest", lowess_frac=0.3):
    # Output diagnostics of a sweep of burnins.
    # Use when you want to see how the burnins are sweeping across outputs like cases/EIR/immunity, and
    # identify most "representative" burnins for each transmission intensity level
    #
    # Pseudocode:
    # - Get endpoints from last few years of sims
    # - Calculate distance of each burnin from the "middle", as defined by input agg
    # - Save figures showing this information.

    def _add_smoothed_values(df_burnin):
        df_burnin.sort_values(by=[tags_to_group_by, "Run_Number"], inplace=True, ignore_index=True)

        for c in columns_of_interest:
            x = df_burnin[tags_to_group_by]
            y = df_burnin[c]
            xs, ys = fit_lowess_spline(x, y, frac=lowess_frac)

            # Find any negative values and replace them with mean of actual burnin:
            needs_correction = ys < 0
            corrected = np.mean(y[needs_correction])
            ys[needs_correction] = corrected

            df_burnin[c + "_SMOOTHED"] = ys

        return df_burnin

    def _plot_data(df_burnin, min_dist_df):
        # Plot burnins:
        plt.close('all')
        plt.figure(figsize=(10, 10), dpi=300)

        x = df_burnin[tags_to_group_by]


        i = 1
        for c in columns_of_interest:
            y = df_burnin[c]
            ys = df_burnin[c + "_SMOOTHED"]

            plt.subplot(3, 2, i)
            plt.scatter(x, y, zorder=0)
            plt.plot(x, ys, c="C1", zorder=1)

            x_selected = min_dist_df[tags_to_group_by]
            y_selected = min_dist_df[c]
            plt.scatter(x_selected, y_selected, marker='*', zorder=2)

            plt.ylabel(c)
            plt.xlabel(tags_to_group_by)

            i += 1

        plt.tight_layout()
        # plt.show()
        plt.savefig("burnin_endpoints.png")

    def _find_most_representative_burnins(df_burnin):
        for c in columns_of_interest:
            y = df_burnin[c]
            ys = df_burnin[c + "_SMOOTHED"]

            # distance
            d = np.abs(y-ys)**2/ys**2
            df_burnin[c+ "_DISTANCE"] = d

        for i in range(len(columns_of_interest)):
            if i == 0:
                df_burnin["TOTAL_DISTANCE"] = df_burnin[columns_of_interest[i] + "_DISTANCE"]
            else:
                df_burnin["TOTAL_DISTANCE"] = df_burnin["TOTAL_DISTANCE"] + df_burnin[columns_of_interest[i] + "_DISTANCE"]


        # Loop over dataframe (bad?) and find smallest distance burnin for each value of the transmission tag
        dict_list = []
        for i, sdf in df_burnin.groupby(tags_to_group_by):
            min_dist_row_as_dict = dict(sdf.loc[sdf["TOTAL_DISTANCE"].idxmin()])
            dict_list.append(min_dist_row_as_dict.copy())

        min_dist_df = pd.DataFrame(dict_list)
        return min_dist_df

    def _add_outpath_to_df(min_dist_df):
        # create sim map, and merge
        sim_map = get_sim_map(experiment_id)
        df_return = pd.merge(min_dist_df, sim_map[["sim_id", "outpath"]], how="left", on="sim_id")
        return df_return


    # =============================================================

    # if csv_filename is not None:
    #     df_burnin = pd.read_csv(csv_filename)
    # elif os.path.isfile("burnin_map.csv"):
    #     df_burnin = pd.read_csv("burnin_map.csv")
    # else:
    #     print("Need burnin_map.csv")
    #     #fixme Run analyzer that outputs this:
    #     # run_analyzer_as_ssmt(experiment_id=experiment_id, analyzers=[SimEndpoint], analyzer_args=[{"save_file": "test.csv"}])


    columns_of_interest = ["aeir",
                           "annual_cases_per_1000",
                           "annual_rdt_prev",
                           "pfemp_frac",
                           "annual_reported_cases_per_1000",
                           "pfpr_2-10"]

    # Omit columns that aren't available:
    columns_of_interest = [c for c in columns_of_interest if c in list(df_burnin.columns)]

    df_burnin = _add_smoothed_values(df_burnin)
    min_dist_df = _find_most_representative_burnins(df_burnin)
    _plot_data(df_burnin, min_dist_df)

    # Finally, create CSV with output directory of most representative burnin for each transmission tag
    # df_save = _add_outpath_to_df(min_dist_df)
    # df_save = df_save[[transmission_tag, "outpath"]]
    # if experiment_id is not None:
    df_final = _add_outpath_to_df(min_dist_df)
    # else:
    #     df_final = min_dist_df.copy()
    smoothed_columns_of_interest = [x+"_SMOOTHED" for x in columns_of_interest]
    df_save = df_final[[tags_to_group_by, "sim_id", "outpath"] + smoothed_columns_of_interest]

    df_save.to_csv("burnin_outpaths.csv", index=False)

    # for s in min_dist_df["sim_id"]:

    # def process_outpath(sim):
    #     outpath = sim.get_platform_object().hpc_jobs[0].working_directory.replace('internal.idm.ctr', 'mnt')
    #     outpath = outpath.replace('\\', '/')
    #     outpath = outpath.replace('IDM2', 'idm2')
    #     return outpath
    #
    # def get_sim_map(platform, burnin_exp_id):
    #     exp = Experiment.from_id(burnin_exp_id, children=False)
    #     exp.simulations = platform.get_children(exp.id, exp.item_type,
    #                                             children=["tags", "configuration", "files", "hpc_jobs"])
    #
    #     dict_list = []
    #
    #     for simulation in exp.simulations:
    #         sim_data_dict = copy(dict(simulation.tags))
    #         sim_data_dict["outpath"] = process_outpath(simulation)
    #
    #         dict_list.append(sim_data_dict)
    #
    #
    #     df = pd.DataFrame(dict_list)
    #     return df








# def find_representative_seed(exp_id, agg="median", on="funest", savefig=False,


if __name__=="__main__":
    # survey_burnin_sweep(experiment_id="5e7e2505-9a5f-ec11-a9f1-9440c9be2c51", tags_to_group_by="larval_habitat_scale")
    survey_burnin_sweep(experiment_id="8e46e3ff-e339-ed11-a9fc-b88303911bc1", tags_to_group_by="habitat_scale")