# Originally from Caitlin Bever
# Functions for converting from simulation day to real date, or vice versa

import numpy as np
import pandas as pd

from datetime import *
from dateutil import relativedelta


def convert_to_day(convert_date, ref_date, date_format="%Y-%m-%d"):
    # Converts date to day of simulation starting from reference date
    # Uses actual calendar dates
    date_delta = datetime.strptime(convert_date, date_format) - \
                 datetime.strptime(ref_date, date_format)

    return date_delta.days


def convert_to_date(convert_day, ref_date, date_format="%Y-%m-%d"):
    # Converts day of simulation starting from reference date into date
    # Uses actual calendar dates
    full_date = datetime.strptime(ref_date, date_format) + relativedelta.relativedelta(days=int(convert_day))

    return datetime.strftime(full_date, date_format)


def convert_to_day_365(convert_date, ref_date, date_format="%Y-%m-%d"):
    # Converts date to day of simulation starting from reference date
    # Assumes a calendar year has exactly 365 days
    return 365 * (datetime.strptime(convert_date, date_format).year - \
                  datetime.strptime(ref_date, date_format).year) + \
           datetime.strptime(convert_date, date_format).timetuple().tm_yday


def convert_to_date_365(convert_day, ref_date, date_format="%Y-%m-%d"):
    # Converts day of simulation starting from reference date into date
    # Assumes a calendar year has exactly 365 days

    convert_year = int(int(convert_day) / 365) + datetime.strptime(ref_date, date_format).year
    convert_day = convert_day - int(int(convert_day) / 365) * 365
    return datetime.strftime(datetime(convert_year, 1, 1) + timedelta(convert_day - 1), date_format)


def day_number_to_month():
    day_to_month = [1] * 31 + [2] * 28 + [3] * 31 + [4] * 30 + [5] * 31 + [6] * 30 \
                   + [7] * 31 + [8] * 31 + [9] * 30 + [10] * 31 + [11] * 30 + [12] * 31
    return day_to_month

def map_day_number_to_month(day_nums):
    df_input = pd.DataFrame({
        "day": day_nums
    },
    index=np.arange(len(day_nums)))

    df_mapping = pd.DataFrame({
        "day": list(range(1,365+1)),
        "month": day_number_to_month()
    })

    df_output = pd.merge(df_input, df_mapping, how="left", on="day")
    return list(df_output["month"])


def week_number_to_month(wn):
    # Returns dictionary which gives fraction of week in each month:

    month_dict = {
        1: 0,
        2: 0,
        3: 0,
        4: 0,
        5: 0,
        6: 0,
        7: 0,
        8: 0,
        9: 0,
        10: 0,
        11: 0,
        12: 0,
    }

    if wn <= 4:
        month_dict[1] = 1
    elif wn == 5:
        month_dict[1] = 3 / 7
        month_dict[2] = 4 / 7
    elif wn >= 6 and wn <= 8:
        month_dict[2] = 1
    elif wn == 9:
        month_dict[2] = 3 / 7
        month_dict[3] = 4 / 7
    elif wn >= 10 and wn <= 12:
        month_dict[3] = 1
    elif wn == 13:
        month_dict[3] = 6 / 7
        month_dict[4] = 1 / 7
    elif wn >= 14 and wn <= 17:
        month_dict[4] = 1
    elif wn == 18:
        month_dict[4] = 1 / 7
        month_dict[5] = 6 / 7
    elif wn >= 19 and wn <= 21:
        month_dict[5] = 1
    elif wn == 22:
        month_dict[5] = 4 / 7
        month_dict[6] = 3 / 7
    elif wn >= 23 and wn <= 25:
        month_dict[6] = 1
    elif wn == 26:
        month_dict[6] = 6 / 7
        month_dict[7] = 1 / 7
    elif wn >= 27 and wn <= 30:
        month_dict[7] = 1
    elif wn == 31:
        month_dict[7] = 2 / 7
        month_dict[8] = 5 / 7
    elif wn >= 32 and wn <= 34:
        month_dict[8] = 1
    elif wn == 35:
        month_dict[8] = 5 / 7
        month_dict[9] = 2 / 7
    elif wn >= 36 and wn <= 39:
        month_dict[9] = 1
    elif wn >= 40 and wn <= 43:
        month_dict[10] = 1
    elif wn == 44:
        month_dict[10] = 3 / 7
        month_dict[11] = 4 / 7
    elif wn >= 45 and wn <= 47:
        month_dict[11] = 1
    elif wn == 48:
        month_dict[11] = 5 / 7
        month_dict[12] = 2 / 7
    else:
        month_dict[12] = 1

    return month_dict


def aggregate_weeks_to_months(df):
    # Assumes df has fields "period" in the format "2015W50", and "value"
    # Creates a new dataframe with columns "year", "month", and "value"
    # Where value is the weekly data now aggregated to the monthly level

    return_dict = {}
    for index, row in df.iterrows():
        week_str = row["period"]
        year_str = week_str.split('W')[0]
        week_num = np.int32(week_str.split('W')[1])

        month_dict = week_number_to_month(week_num)

        for month in range(1, 13):
            month_str = year_str + str(int(month)).zfill(2)

            val = month_dict[month] * row["value"]

            if month_str in return_dict:
                return_dict[month_str] = return_dict[month_str] + val
            else:
                return_dict[month_str] = val

    reorg_dict = {"period": list(return_dict.keys()), "value": list(return_dict.values())}
    return pd.DataFrame(reorg_dict)


# Approximate day number for each month
month_times = [
    0.0,
    30.417,
    60.833,
    91.25,
    121.667,
    152.083,
    182.5,
    212.917,
    243.333,
    273.75,
    304.167,
    334.583
]

def bin_daily_data_by_month(df, day_column_name, data_column_name, agg_func="mean"):
    day_nums = df[day_column_name]

    df["year_num"] = (df[day_column_name]/365).astype("int")
    df["day_number_in_year"] = day_nums - df["year_num"]*365
    df["month_number_in_year"] = map_day_number_to_month(list(df["day_number_in_year"]))
    df["month_num"] = 12*df["year_num"] + df["month_number_in_year"]

    # if agg_func == "mean":
    df_output = df.groupby("month_num").agg({data_column_name: agg_func}).reset_index()
    return df_output
    # if agg_func != "mean":
    #     raise NotImplementedError
